@@ -8,8 +8,9 @@ from rest_framework.authentication import BaseAuthentication
 from rest_framework.permissions import BasePermission
 from rest_framework.request import Request
 from rest_framework.status import HTTP_200_OK, HTTP_201_CREATED, HTTP_204_NO_CONTENT, HTTP_404_NOT_FOUND, \
-    HTTP_400_BAD_REQUEST
+    HTTP_400_BAD_REQUEST, HTTP_500_INTERNAL_SERVER_ERROR
 from sqlalchemy.exc import StatementError
+from sqlalchemy.inspection import inspect
 from sqlalchemy.orm import Query
 from sqlalchemy_filters import apply_filters
 from sqlalchemy_filters import apply_pagination
@@ -30,9 +31,10 @@ class AlchemyJsonApiViewBuilder:
                  alchemy_model: Type,
                  resource_name: str,
                  fields: Sequence[str],
+                 api_version: Optional[str] = '',
                  primary_key: Optional[str] = 'id',
                  allowed_methods: Optional[Sequence[str]] = json_api_spec_http_methods.HTTP_ACTIONS,
-                 base_query: Optional[Query] = None,
+                 base_query: Optional[Callable[[Any], Query]] = None,
                  permitted_objects: Optional[Callable[[Request, Query], Query]] = None,
                  permission_classes: Sequence[Type[BasePermission]] = None,
                  authentication_classes: Sequence[Type[BaseAuthentication]] = None,
@@ -64,6 +66,7 @@ class AlchemyJsonApiViewBuilder:
         self._relations = []
         self._custom_fields = {}
         self._computed_filters = {}
+        self._api_version = api_version
         self._skip_plugins = skip_plugins if skip_plugins is not None else [plugins.AUTO_ADMIN_VIEWS]
         include_plugins = include_plugins or []
         for item in include_plugins:
@@ -127,10 +130,12 @@ class AlchemyJsonApiViewBuilder:
                      model: Type,
                      resource_name: str,
                      many: bool = False,
-                     primary_key_name: str = None) -> 'AlchemyJsonApiViewBuilder':
+                     primary_key_name: str = None,
+                     api_version: Optional[str] = '') -> 'AlchemyJsonApiViewBuilder':
         self._relations.append(AlchemyRelation(field_name=field, model=model, many=many,
                                                resource_name=resource_name,
-                                               primary_key=primary_key_name))
+                                               primary_key=primary_key_name,
+                                               api_version=api_version))
         return self
 
     def add_computed_filter(self, name: str,
@@ -147,7 +152,7 @@ class AlchemyJsonApiViewBuilder:
         admin_builder = deepcopy(self)
         admin_builder._skip_plugins = [plugins.AUTO_ADMIN_VIEWS]
         admin_builder._permitted_objects = None
-        admin_builder._resource_name = f'admin_view_{admin_builder._resource_name}'
+        admin_builder._resource_name = f'admin_view_{admin_builder._resource_name.replace("admin_view_", "")}'
         admin_permission_class = admin_builder._plugin_options.get(plugins.AUTO_ADMIN_VIEWS, {}).get(
             'ADMIN_PERMISSION_CLASS')
 
@@ -187,7 +192,7 @@ class AlchemyJsonApiViewBuilder:
                 if include_on_model:
                     model_property = include_on_model.property
                     #  Get the primary key of the target relationship table
-                    target_primary_key = model_property.target.primary_key.columns.values()[0].name
+                    primary_join = model_property.primaryjoin
                     #  Get the local foreign key column name that connects the relationship
                     local_column = model_property.local_columns.copy().pop()
                     #  Get the Alchemy Model of the target relationship table
@@ -195,10 +200,13 @@ class AlchemyJsonApiViewBuilder:
                     #  Check that the item we are targeting has a JSON:API view
                     if target_model not in _TYPE_TO_SCHEMA:
                         raise Exception(f'No JSON:API view defined for type {target_model}')
+                    primary_key = inspect(self._model).primary_key[0].name
                     #  Fetch all the related items to be included in the result
-                    to_include = target_model.objects.query().filter(
-                        getattr(target_model, target_primary_key).in_(
-                            [getattr(item, local_column.name) for item in objects])).all()
+                    to_include = target_model.objects.query().filter(primary_join,
+                                                                     getattr(self._model, primary_key).in_(
+                                                                         [getattr(item, local_column.name) for item in
+                                                                          objects]
+                                                                     )).all()
                     schema = _TYPE_TO_SCHEMA[target_model]
                     target_many = schema['serializer'](many=True)
                     #  Serialize all the included objects to JSON:API
@@ -212,7 +220,8 @@ class AlchemyJsonApiViewBuilder:
         permitted_objects = self._permitted_objects or default_permitted_objects
 
         def object_get(request, identifier, *args, **kwargs) -> Tuple[Dict, int]:
-            permitted_query = permitted_objects(request, self._base_query or self._model.objects.query())
+            permitted_query = permitted_objects(request,
+                                                self._base_query() if self._base_query is not None else self._model.objects.query())
 
             obj = None
 
@@ -227,13 +236,12 @@ class AlchemyJsonApiViewBuilder:
             if self._after_get_callback:
                 obj = self._after_get_callback(request, obj)
 
-            result = schema.json_api_dump(obj,  self._resource_name)
-            # attributes.pop(self._primary_key or 'id', None)
+            result = schema.json_api_dump(obj, self._resource_name)
             return result, HTTP_200_OK
-            # return {'id': identifier, 'type': self._resource_name, 'attributes': attributes}, HTTP_200_OK
 
         def object_list(request, page, filters=None, includes=None, *args, **kwargs) -> Tuple[List, List, int, int]:
-            permitted_query = permitted_objects(request, self._base_query or self._model.objects.query())
+            permitted_query = permitted_objects(request,
+                                                self._base_query() if self._base_query is not None else self._model.objects.query())
             #
             #  Apply all the filters from the URL
             #
@@ -253,7 +261,7 @@ class AlchemyJsonApiViewBuilder:
             #
             #  Paginate the result
             #
-            query, pagination = apply_pagination(filtered_query, page_number=page, page_size=self._page_size)
+            query, pagination = apply_pagination(filtered_query, page_number=int(page), page_size=self._page_size)
 
             if self._before_list_callback:
                 query = self._before_list_callback(request, query)
@@ -264,7 +272,10 @@ class AlchemyJsonApiViewBuilder:
             rendered_includes = render_includes(includes, objects)
 
             if self._after_list_callback:
-                objects = self._after_list_callback(request, objects)
+                try:
+                    objects = self._after_list_callback(request, objects)
+                except Exception as e:
+                    return [{'errors': [str(e)]}], [], 0, getattr(e, 'http_status', HTTP_500_INTERNAL_SERVER_ERROR)
 
             result = schema_many.json_api_dump(objects, self._resource_name)
             return result, rendered_includes, pagination.total_results, HTTP_200_OK
@@ -275,7 +286,10 @@ class AlchemyJsonApiViewBuilder:
             else:
                 attributes = data['attributes']
             if self._before_create_callback:
-                attributes = self._before_create_callback(request, attributes)
+                try:
+                    attributes = self._before_create_callback(request, attributes)
+                except Exception as e:
+                    return {'errors': [str(e)]}, '', getattr(e, 'http_status', HTTP_500_INTERNAL_SERVER_ERROR)
 
             try:
                 unmarshal_obj = schema.load(attributes, session=schema.db.session, unknown=INCLUDE)
@@ -286,17 +300,18 @@ class AlchemyJsonApiViewBuilder:
             obj.save()
 
             if self._after_create_callback:
-                self._after_create_callback(request, attributes, obj)
+                try:
+                    self._after_create_callback(request, attributes, obj)
+                except Exception as e:
+                    return {'errors': [str(e)]}, '', getattr(e, 'http_status', HTTP_500_INTERNAL_SERVER_ERROR)
                 obj.refresh_from_db()
 
-            dumped_attributes = schema.dump(obj)
-            dumped_attributes.pop(self._primary_key or 'id', None)
-
-            return {'data': {'type': self._resource_name, 'id': obj.id,
-                             'attributes': dumped_attributes}}, obj.id, HTTP_201_CREATED
+            result = schema.json_api_dump(obj, self._resource_name)
+            return result, obj.id, HTTP_201_CREATED
 
         def object_update(request, identifier, data, *args, **kwargs) -> Tuple[Dict, int]:
-            permitted_query = permitted_objects(request, self._base_query or self._model.objects.query())
+            permitted_query = permitted_objects(request,
+                                                self._base_query() if self._base_query is not None else self._model.objects.query())
             obj = None
             try:
                 obj = permitted_query.filter_by(**{self._primary_key or 'id': identifier}).first()
@@ -308,7 +323,10 @@ class AlchemyJsonApiViewBuilder:
 
             attributes = data['attributes']
             if self._before_update_callback:
-                attributes = self._before_update_callback(request, attributes, obj)
+                try:
+                    attributes = self._before_update_callback(request, attributes, obj)
+                except Exception as e:
+                    return {'errors': [str(e)]}, getattr(e, 'http_status', HTTP_500_INTERNAL_SERVER_ERROR)
                 obj.refresh_from_db()
 
             for field in self._fields:
@@ -317,28 +335,36 @@ class AlchemyJsonApiViewBuilder:
             obj.save()
 
             if self._after_update_callback:
-                self._after_update_callback(request, obj)
+                try:
+                    self._after_update_callback(request, obj)
+                except Exception as e:
+                    return {'errors': [str(e)]}, getattr(e, 'http_status', HTTP_500_INTERNAL_SERVER_ERROR)
                 obj.refresh_from_db()
 
-            dumped_attributes = schema.dump(obj)
-            dumped_attributes.pop(self._primary_key or 'id', None)
-
-            return {'data': {'type': self._resource_name, 'id': identifier,
-                             'attributes': dumped_attributes}}, HTTP_200_OK
+            result = schema.json_api_dump(obj, self._resource_name)
+            return result, HTTP_200_OK
 
         def object_delete(request, identifier, *args, **kwargs) -> Tuple[int]:
-            permitted_query = permitted_objects(request, self._base_query or self._model.objects.query())
+            permitted_query = permitted_objects(request,
+                                                self._base_query() if self._base_query is not None else self._model.objects.query())
             obj = permitted_query.filter_by(**{self._primary_key or 'id': identifier}).first()
             if self._before_delete_callback:
-                obj = self._before_delete_callback(request, obj)
+                try:
+                    obj = self._before_delete_callback(request, obj)
+                except Exception as e:
+                    return getattr(e, 'http_status', HTTP_500_INTERNAL_SERVER_ERROR)
             if obj:
                 obj.delete()
 
                 if self._after_delete_callback:
-                    self._after_delete_callback(request, obj)
+                    try:
+                        self._after_delete_callback(request, obj)
+                    except Exception as e:
+                        return getattr(e, 'http_status', HTTP_500_INTERNAL_SERVER_ERROR)
             return HTTP_204_NO_CONTENT
 
         builder = JsonApiResourceViewBuilder(action_name=self._resource_name,
+                                             api_version=self._api_version,
                                              unique_identifier=self._primary_key,
                                              allowed_methods=self._allowed_methods,
                                              permission_classes=self._permission_classes,
